import cv2
import numpy as np
from rtree import index
from PIL import Image

from dataclasses import dataclass
from typing import List

from tocky.utils import Rect

# This one is all ChatGPT :grimacing: Should verify
# Calculate the shortest distance between the two rectangles
def shortest_distance(rect1: Rect, rect2: Rect):
    # Calculate the distance between the right edge of rect1 and the left edge of rect2
    dx = max(0, rect2.left - rect1.right, rect1.left - rect2.right)

    # Calculate the distance between the bottom edge of rect1 and the top edge of rect2
    dy = max(0, rect2.top - rect1.bottom, rect1.top - rect2.bottom)

    # Return the Euclidean distance between the two rectangles
    return (dx**2 + dy**2)**0.5

def in_to_px(inches: float, dpi: int) -> float:
  return inches * dpi

def px_to_in(pixels: float, dpi: int) -> float:
  return pixels / dpi

@dataclass
class LeaderDotsParams:
  dpi: int
  # Note these are in inches -- we represent them in inches to try to make this
  # more resilient to changes in DPI
  _max_dot_height = 0.05
  _kernel_size = 0.015
  _dot_isolation_distance = 0.04
  _inpaint_radius = 0.04

  @property
  def max_dot_height(self) -> int: return round(in_to_px(self._max_dot_height, self.dpi))
  @property
  def kernel_size(self) -> int: return round(in_to_px(self._kernel_size, self.dpi))
  @property
  def dot_isolation_distance(self) -> int: return round(in_to_px(self._dot_isolation_distance, self.dpi))
  @property
  def inpaint_radius(self) -> int: return round(in_to_px(self._inpaint_radius, self.dpi))


def remove_leader_dots(img: Image.Image, dpi: int = 300) -> Image.Image:
  P = LeaderDotsParams(dpi)

  # Load your scanned image
  image = np.array(img.convert('RGB'))

  gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)

  # Performing OTSU threshold
  ret, thresh1 = cv2.threshold(gray, 0, 255, cv2.THRESH_OTSU | cv2.THRESH_BINARY_INV)

  # Specify structure shape and kernel size.
  # Kernel size increases or decreases the area
  # of the rectangle to be detected.
  # A smaller value like (10, 10) will detect
  # each word instead of a sentence.
  rect_kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (P.kernel_size, P.kernel_size))

  # Applying dilation on the threshold image
  dilation = cv2.dilate(thresh1, rect_kernel, iterations = 1)

  # Finding contours
  contours, hierarchy = cv2.findContours(dilation, cv2.RETR_EXTERNAL,
                                                  cv2.CHAIN_APPROX_NONE)

  # Create an rtree index
  rect_idx = index.Index(properties=index.Property())
  contour_rects: List[Rect] = []

  # Insert bounding boxes of contours into the R-tree index
  for i, contour in enumerate(contours):
      rect = Rect.from_xywh(cv2.boundingRect(contour))
      contour_rects.append(rect)
      rect_idx.insert(i, rect.to_ltrb())

  # Creating a copy of image
  # im2 = image.copy()
  mask = np.zeros_like(image, dtype=np.uint8)
  mask = cv2.cvtColor(mask, cv2.COLOR_BGR2GRAY)
  mask = mask.astype(np.uint8)

  for i, rect in enumerate(contour_rects):
      # Drawing a rectangle on copied image
      if rect.height < P.max_dot_height:
        closest_rect_indices = [
          result
          for result in rect_idx.nearest(rect.to_ltrb(), 7)
          if result != i
        ]
        closest_rects = [contour_rects[i] for i in closest_rect_indices]
        closest_rect_dist = closest_rects[0]
        closest_dist = shortest_distance(closest_rects[0], rect)
        erase = (
            (
                (closest_dist > P.dot_isolation_distance) and
              len([r for r in closest_rects if r.height < P.max_dot_height]) >= 2
            ) or shortest_distance(closest_rects[0], rect) >= 25
        )
        if erase:
          # cv2.rectangle(im2, (rect.x, rect.y), (rect.right, rect.bottom), (255, 255, 255), thickness=cv2.FILLED)
          cv2.rectangle(mask, (rect.x, rect.y), (rect.right, rect.bottom), (255, 255, 255), thickness=cv2.FILLED)

  reference_area = cv2.ximgproc.guidedFilter(image, mask, radius=P.inpaint_radius, eps=1.0)
  result = cv2.inpaint(image, mask, inpaintRadius=P.inpaint_radius, flags=cv2.INPAINT_TELEA)

  return Image.fromarray(result)



from dataclasses import dataclass, field
from typing import Literal, Any

from tocky.detector.ocr_detector import OcrDetector, extract_leaf_num, ocaid_to_djvu_url

@dataclass
class TestCase:
  ocaid: str
  toc_pages: list[str]
  custom_labels: list[str] = field(default_factory=list)

  @property
  def labels(self):
    labels = list(self.custom_labels)
    if len(self.toc_pages) == 0:
      labels.append('No TOC')
    elif len(self.toc_pages) == 1:
      labels.append('Single Page TOC')
    else:
      labels.append('Multi Page TOC')

    return labels

test_set = [
    TestCase('countofmontecris01duma', ['0013'], ['Has List of Illustrations']),
    TestCase('travelsinmexicod01gill', []),
    TestCase('davidpoindexters00hawt', ['0007']),
    TestCase('harvardclassicss18elio', ['0005', '0006']),
    TestCase('prideprejudice0000unse_r8e0', [], ['Has List of Illustrations']),
    TestCase('englishstruwwelp00hoffrich', []),
    TestCase('janeeyreautobiog01bron', []),
    TestCase('cu31924013243963', ['0012', '0013'], ['Has List of Illustrations']),
    TestCase('adventuresofsher00doylrich', ['0011'], ['Has List of Illustrations', 'Contents page has bad OCR']),
    TestCase('cihm_93635', [], ['Microfilm']),
    TestCase('murderofrogerack0000agat_g7y3', ['0011'], ['Printing history on copyright page']),
    TestCase('alicesadventures0000lewi_h0b9', ['0015']),
    TestCase('dracu00stok', ['0013', '0014', '0015'], ['page numbers are not separate words']),
    TestCase('odysseybookiv00home', [], ['bad OCR on blank page']),
    TestCase('the-great-gatsby', []),
    TestCase('olivertwist0000char_v6j2', ['0015', '0016', '0017']),
    TestCase('littleprincessbe00burn', ['0017'], ['Has List of Illustrations']),
    TestCase('alicesadventures0000unse_r2e4', ['0009', '0010'], ['Contents page has bad OCR']),
    TestCase('dianearbuss1960s0000gros', ['0009']),
    TestCase('meditationsofmar00marc', ['0009']),
    TestCase('kingjamesfirstdm00jame', []),
    TestCase('kamasutraofvatsy00vatsuoft', ['0011']),
    TestCase('gameoflifehowtop00shin', ['0005']),
    TestCase('alanturingenigma0000hodg', ['0011']),
    # Super slow
    # TestCase('dictionaryofprin00drozrich', [], ['bad OCR on blank page']),
    TestCase('b24884431', [], ['Has a table in front matter']),
    TestCase('artofpersuasionw0000burg', ['0011', '0012'], ['Contents page has bad OCR']),
    TestCase('atomichabitseasy0000clea', ['0009', '0010', '0011'], ['bad OCR on blank page']),
    TestCase('dressmakingselft00care', ['0009', '0010', '0011', '0012', '0013'], ['Sparse TOC pages']),
    TestCase('whoswhoincanada0072unse', [], ['Index at front']),
    TestCase('20yearsevergood0000unse', ['0013'], ['Weird TOC format']),
    TestCase('isbn_9780039105167', ['0007', '0008'], ['Contents page has bad OCR']),
    TestCase('b2892907x', [], ['Has a table in front matter']),
    TestCase('civilwarnarrativ0000foot_w4t8', ['0011'], ['bad OCR on blank page']),
    TestCase('pilgrimmemorials00russ_0', [], ['Index at front']),
    TestCase('agriculturalpoli0000elli', ['0009', '0010', '0011', '0012'], ['Dense TOC']),
    TestCase('leipzeigtrialsac00mull', ['0025'], ['TOC after page 20']),
    TestCase('coloradosrodeoro0000ordw', ['0009'], ['Contents page has bad OCR']),
    TestCase('dirtylaundry0000cull', ['0009'], ['Contents page has bad OCR']),
    TestCase('employmentlaw06edcarr', ['0005', '0006', '0007']),
    TestCase('solarcellstheira0000unse', [f'{d:04}' for d in range(9, 20)], ['Contents page has bad OCR']),
    TestCase('warsoffrenchdeco0000clay', ['0007', '0008', '0009']),
    TestCase('piepiepieeasyhom0000carr', ['0008', '0009'], ['Page numbers on left of TOC entries']),
    TestCase('essentialspanish00nova', ['0009', '0010', '0011', '0012', '0013', '0014', '0015']),
    TestCase('davidsonsessenti0000unse_k5e3', ['0009'], ['Contents page has bad OCR']),
    TestCase('whyscientistsdis0000unse', ['0011'], ['Contents page has bad OCR']),
    TestCase('plowsplantingimp0000halb', ['0007']),
    TestCase('phonicsphonemica0000cala', ['0004']),
    TestCase('writingtohealgui0000penn', ['0007', '0008']),
    TestCase('prayersfromseaso0000deck', ['0007', '0008'], ['Contents page OCR missing numbers']),
    TestCase('artbasedgames0000pave', ['0007', '0008']),
    TestCase('transitionfromsc0000unse', ['0007', '0008', '0009']),
    TestCase('workbookofphotog0000hedg', ['0006', '0007'], ['Many entires']),
    TestCase('subtleartofnotgi0000mans_n2n5', ['0005', '0006', '0007']),
    TestCase('digitalphotograp00shep', [f'{d:04}' for d in range(8, 13)])
    #TestCase('', []),
]

@dataclass
class TestResult:
  test: TestCase
  status: Literal['PASS',  'FAIL', 'ERRO']
  actual: Any
  reasons: Any

def run_test(test: TestCase) -> TestResult:
  expected_pages = [int(p) for p in test.toc_pages]
  pages_to_see = set(expected_pages)
  test_passes = True
  reasons = {}
  has_begun = False
  has_ended = False
  toc_pagenums = []
  for (pagenum, xml_str, toc_analysis) in OcrDetector().analyze_djvu_for_toc(ocaid_to_djvu_url(test.ocaid)):
    pagen = extract_leaf_num(pagenum)

    has_begun = has_begun or toc_analysis.is_toc
    is_expected = pagen in expected_pages

    if toc_analysis.is_toc:
      toc_pagenums.append(pagen)
    if has_begun and not toc_analysis.is_toc:
      has_ended = True

    if toc_analysis.is_toc == is_expected == True:
      # We good!
      pass
    elif not toc_analysis.is_toc and is_expected:
      test_passes = False
      reasons[pagen] = toc_analysis.failure
    elif toc_analysis.is_toc and not is_expected:
      test_passes = False
      reasons[pagen] = 'Wrong TOC page'

    if is_expected:
      pages_to_see.remove(pagen)
    if has_ended and not pages_to_see:
      break

  result = [f'{n:04}' for n in toc_pagenums]
  return TestResult(test, 'PASS' if result == test.toc_pages else 'FAIL', result, reasons)

def run_all_tests(test_set: list[TestCase]):
  import concurrent.futures

  icon_mapping = {
    'PASS': '✔️',
    'FAIL': '❌',
    'ERRO': '💥',
  }

  def process_test(test):
    result = run_test(test)
    icon = icon_mapping[result.status]
    print(f'{icon} {result.test.ocaid}')
    if result.status != 'PASS':
        print('Expected:', result.test.toc_pages)
        print('Actual:', result.actual)
        print('Links:', [f"https://archive.org/details/{result.test.ocaid}/page/leaf{n}/mode/1up" for n in result.actual])
        print('Labels:', result.test.labels)
        print('Reasons:', result.reasons)

    return result

  with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
      return list(executor.map(process_test, test_set))



from dataclasses import dataclass
import math
import openai

import tiktoken


@dataclass
class PriceRangeCents:
    min: float
    max: float

    def __repr__(self):
        return f"{self.min / 100:.4f} - {self.max / 100:.4f} USD"

@dataclass
class ModelPricing:
    name: str
    prompt_cents: int
    """Cost in cents per 1,000,000 tokens"""
    completion_cents: int
    """Cost in cents per 1,000,000 tokens"""
    image_base_tokens: int
    """Base token cost for an image"""
    image_tile_size: int
    """Size of a square image tile in pixels"""
    tokens_per_tile: int
    """Tokens per image tile"""

    def cost_from_usage(self, usage: openai.types.CompletionUsage):
        return self.cost(usage.prompt_tokens, usage.completion_tokens)

    def cost(self, prompt_tokens: int, completion_tokens: int):
        prompt_cost = prompt_tokens * self.prompt_cents / 1_000_000
        completion_cost = completion_tokens * self.completion_cents / 1_000_000
        return prompt_cost + completion_cost

    def count_tokens(self, message: str) -> int:
        return len(tiktoken.encoding_for_model(self.name).encode(message))

    def predict_cost(
        self,
        prompts: list[str],
        max_tokens: int,
        image_sizes: list[tuple[int, int]],
    ):       
        prompt_tokens = sum(self.count_tokens(p) for p in prompts)
        image_tokens = sum(
            self.tokens_per_tile * math.ceil(image_size[0] / self.image_tile_size) * math.ceil(image_size[1] / self.image_tile_size)
            for image_size in image_sizes
        )
        if image_tokens:
            image_tokens += self.image_base_tokens
        
        prompt_cost = (prompt_tokens + image_tokens) * self.prompt_cents / 1_000_000
        completion_cost = max_tokens * self.completion_cents / 1_000_000

        return PriceRangeCents(min=prompt_cost, max=prompt_cost + completion_cost)

MODEL_PRICES = {
    "gpt-4o-mini": ModelPricing(
        name="gpt-4o-mini",
        prompt_cents=15,
        completion_cents=60,
        image_base_tokens=2833,
        image_tile_size=512,
        tokens_per_tile=5667,
    ),
}

def predict_cost(
    model_name: str,
    prompts: list[str],
    max_tokens: int,
    image_sizes: list[tuple[int, int]],
):
    return MODEL_PRICES[model_name].predict_cost(prompts, max_tokens, image_sizes)



from dataclasses import dataclass, field
import dataclasses
from typing import Literal, TypedDict
import json
import traceback
import requests


from tocky import DETECTERS_BY_NAME, EXTRACTORS_BY_NAME
from tocky.detector import AbstractDetector
from tocky.detector.ocr_detector import OcrDetector
from tocky.env import get_env
from tocky.extractor import AbstractExtractor, TocEntry
from tocky.extractor.ai_extractor import AiExtractor
from tocky.utils.ia import bulk_ia_to_ol, get_ia_metadata
from tocky.utils import ResultStat, get_git_sha, get_tocky_version, run_with_result_stats
from tocky.validator import validate_extracted_toc

TockyItemState = Literal[
  "To Detect",
  "Detecting",
  "To Extract",
  "Extracting",
  "To Review",
  "Reviewing",
  "Done",
  "Errored",
]


# Keep in sync with list.html
ExtractionStatus = Literal[
  "",
  "TOC Extracted",
  "No TOC detected",
  "Errored",
  "Already has good TOC",
  "TOC Validation: Unparseable TOC",
  "TOC Validation: Label/Title Too Long",
  "TOC Validation: No Numbers",
  "TOC Validation: Not Enough Numbers",
  "TOC Validation: Too Few Lines With Numbers",
  "TOC Validation: Starts Too High",
  "TOC Validation: End Too Low",
]

@dataclass
class ItemProcessingState:
  ocaid: str
  detector: AbstractDetector
  extractor: AbstractExtractor

  detector_result: ResultStat[list[int]] | None = None
  extractor_result: ResultStat[list[TocEntry]] | None = None

  state: TockyItemState = 'To Detect'
  status: ExtractionStatus = ''
  toc_raw_ocr: list[str] | None = None
  detected_toc: list[tuple[str, str]] = field(default_factory=list)
  structured_toc: list[TocEntry] | None = None
  error: Exception | None = None
  prompt_tokens: int = 0
  completion_tokens: int = 0

  def to_response_dict(self):
    result = {
        'tocky_version': get_tocky_version(),
        'success': all(
            result and result.success
            for result in [self.detector_result, self.extractor_result]
        ),
        'input_book': {
          'type': 'ia_book',
          'ia_id': self.ocaid,
        },
        'detector': {
            'type': self.detector.name,
            'options': self.detector.P.__dict__,
            'results': self.detector_result.to_dict() if self.detector_result else None,
        },
        'extractor': {
            'type': self.extractor.name,
            'options': self.extractor.P.__dict__,
            'results': self.extractor_result.to_dict() if self.extractor_result else None,
        },
    }

    # Convert TocEntry to dict extractor result
    if self.extractor_result and self.extractor_result.result is not None:
      result['extractor']['results']['result'] = [
          entry.to_dict()
          for entry in self.extractor_result.result
      ]

    return result

  def to_db_dict(self):
    return {
      'state': self.state,
      'status': self.status,

      **self.to_response_dict(),

      # Legacy?
      'code_version': 'v2.E.1',
      'ocaid': self.ocaid,
      'prompt_tokens': self.prompt_tokens,
      'completion_tokens': self.completion_tokens,
      'error': str(self.error) if self.error else None,
      'toc_raw_ocr': self.toc_raw_ocr,
      'structured_toc': [entry.to_dict() for entry in self.structured_toc] if self.structured_toc is not None else None,
      'detected_toc': self.detector_result.result if self.detector_result else None,
  }


def process_ol_book(
  ol_record: dict, 
  detector: AbstractDetector,
  extractor: AbstractExtractor,
  push: bool = False,
) -> ItemProcessingState:
  if ol_toc := ol_record.get('table_of_contents'):
    toc_missing_pagenums = not any(chapter.get('pagenum') for chapter in ol_toc)
    if not toc_missing_pagenums:
      return ItemProcessingState(
        ocaid=ol_record['ocaid'],
        detector=detector,
        extractor=extractor,
        state='Done',
        status='Already has good TOC'
      )
  
  return process_ia_book(ol_record['ocaid'], detector, extractor, push=push)


def process_ia_book(
  ocaid: str,
  detector: AbstractDetector,
  extractor: AbstractExtractor,
  push: bool = False,
) -> ItemProcessingState:
  state = ItemProcessingState(ocaid=ocaid, detector=detector, extractor=extractor)
  toc_queue_id = 0
  if push:
    toc_queue_id = push_to_toc_queue(state.to_db_dict())
  
  def set_state(new_state: TockyItemState):
    state.state = new_state
    if push:
      update_toc_queue(toc_queue_id, state.to_db_dict())

  set_state('Detecting')
  state.detector_result = run_with_result_stats(lambda: detector.detect(state.ocaid))

  if state.detector_result.error is not None:
    state.status = 'Errored'
    state.error = state.detector_result.error
    set_state('Errored')
    return state

  if not state.detector_result.result:
    state.status = 'No TOC detected'
  else:
    set_state('Extracting')
    state.extractor_result = run_with_result_stats(lambda: extractor.extract(state.ocaid, state.detector_result.result))

    state.toc_raw_ocr = extractor.toc_raw_ocr

    if state.extractor_result.error is not None:
      state.status = 'Errored'
      state.error = state.extractor_result.error
      set_state('Errored')
      return state
  

    state.structured_toc = state.extractor_result.result
    assert extractor.toc_response
    state.prompt_tokens = extractor.toc_response.prompt_tokens
    state.completion_tokens = extractor.toc_response.completion_tokens
    state.status = 'TOC Extracted'
    
    if state.structured_toc:
      total_pages = int(get_ia_metadata(state.ocaid)['metadata']['imagecount'])
      validation = validate_extracted_toc(state.structured_toc, total_pages)
      if validation != 'Valid':
        state.status = f'TOC Validation: {validation}'

  set_state('To Review')

  return state

class TockyOptionsError(ValueError):
    pass

def process_from_options(options: dict, push=False):
    if not options['input_book']['ia_id']:
        raise TockyOptionsError('IA ID is required')

    ia_id = options['input_book']['ia_id']

    # Set up detector
    DETECTOR_CLS = DETECTERS_BY_NAME.get(options['detector']['type'])
    if not DETECTOR_CLS:
        raise TockyOptionsError(f'Invalid detector type: {options["detector"]["type"]}')

    detector = DETECTOR_CLS()
    try:
        detector.P = dataclasses.replace(detector.P, **options['detector']['options'])
    except TypeError as e:
        # TODO: This will not error if things are set to the wrong type
      raise TockyOptionsError(f'Invalid detector options: {e}')

    # Run extractor
    EXTRACTOR_CLS = EXTRACTORS_BY_NAME.get(options['extractor']['type'])
    if not EXTRACTOR_CLS:
        raise TockyOptionsError(f'Invalid extractor type: {options["extractor"]["type"]}')

    extractor = EXTRACTOR_CLS()
    try:
        extractor.P = dataclasses.replace(extractor.P, **options['extractor']['options'])
    except TypeError as e:
        # TODO: This will not error if things are set to the wrong type
        raise TockyOptionsError(f'Invalid extractor options: {e}')
    # Share cache
    extractor.S = detector.S

    # Now let's run some stuff!
    detector.debug = False    
    return process_ia_book(ia_id, detector, extractor, push=push)


def push_to_toc_queue(record: dict) -> int:
  resp = requests.put(
      f'{get_env().get_app_prefix()}/push',
      headers={
          'X-API-KEY': get_env().TOCKY_SERVER_KEY,
          'Content-Type': 'application/json',
      },
      data=json.dumps(record)
  ).json()
  return resp['id']

def update_toc_queue(row_id: int, record: dict):
  return requests.post(
      f'{get_env().get_app_prefix()}/update/{row_id}',
      headers={
          'X-API-KEY': get_env().TOCKY_SERVER_KEY,
          'Content-Type': 'application/json',
      },
      data=json.dumps(record)
  ).json()

class IaSearchParams(TypedDict):
  q: str
  sort: str


def process_all(
  ia_params: IaSearchParams,
  rows=10,
  page=1,
  ia_overrides=None,
  push=False,
):
  ia_overrides = ia_overrides or {}
  ia_records = requests.get('https://archive.org/advancedsearch.php', params={
    **ia_params,
    'fl': 'identifier,openlibrary_edition',
    'rows': rows,
    'page': page,
    'output': 'json',
  }).json()['response']['docs']

  for ia_record in ia_records:
    if ia_record['identifier'] in ia_overrides:
      ia_record |= ia_overrides[ia_record['identifier']]

  ol_records_by_key = bulk_ia_to_ol(ia_records)

  import concurrent.futures
  all_results = []

  def run_pipeline(ol_record: dict):
    detector = OcrDetector()
    extractor = AiExtractor()
    return process_ol_book(ol_record, detector, extractor, push=push)

  with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
    for result in executor.map(run_pipeline, ol_records_by_key.values()):
      print(f'[{result.status}] {result.ocaid}')
      push_to_toc_queue(result.to_db_dict())
      if result.error:
        print(traceback.print_exception(result.error))
      all_results.append(result)
  return all_results



import io
from typing import TypedDict, cast
import requests
from tocky.utils import PageScan
from lxml import etree
import os
from PIL import Image

class AzureOcrResponse(TypedDict):
    modelVersion: str
    metadata: 'AzureOcrMetadata'
    readResult: 'AzureOcrReadResult'

class AzureOcrReadResult(TypedDict):
    blocks: list['AzureOcrBlock']

class AzureOcrBlock(TypedDict):
    lines: list['AzureOcrLine']

class AzureOcrMetadata(TypedDict):
    width: int
    height: int

# EG {
#     "text": "9:35 AM",
#     "boundingPolygon": [{"x":131,"y":130},{"x":214,"y":130},{"x":214,"y":148},{"x":131,"y":148}],
#     "words": [{"text":"9:35","boundingPolygon":[{"x":132,"y":130},{"x":172,"y":131},{"x":171,"y":149},{"x":131,"y":148}],"confidence":0.977},{"text":"AM","boundingPolygon":[{"x":180,"y":131},{"x":203,"y":131},{"x":202,"y":149},{"x":180,"y":149}],"confidence":0.998}]
# },
class AzureOcrLine(TypedDict):
    text: str
    boundingPolygon: list['AzureOcrPoint']
    """
    Order of points is top-left, top-right, bottom-right, bottom-left
    """

    words: list['AzureOcrWord']

class AzureOcrPoint(TypedDict):
    x: int
    y: int

class AzureOcrWord(TypedDict):
    text: str
    boundingPolygon: list[AzureOcrPoint]
    confidence: float

def image_to_bytes(image: Image.Image) -> bytes:
    with io.BytesIO() as output:
        image.save(output, format='JPEG')
        return output.getvalue()

def fetch_ocr_azure(image: Image.Image) -> AzureOcrResponse:
    subscription_key = os.getenv("AZURE_SUBSCRIPTION_KEY")
    endpoint = os.getenv("AZURE_ENDPOINT")

    assert subscription_key, "Set the AZURE_SUBSCRIPTION_KEY environment variable."
    assert endpoint, "Set the AZURE_ENDPOINT environment variable."

    endpoint = endpoint.rstrip('/')

    response = requests.post(
        f'{endpoint}/computervision/imageanalysis:analyze',
        params={
            'features': 'read',
            'model-version': 'latest',
            'language': 'en',
            'api-version': '2024-02-01',
        },
        headers={
            'Ocp-Apim-Subscription-Key': subscription_key,
            'Content-Type': 'application/octet-stream'
        },
        data=image_to_bytes(image)
    )
    # response.raise_for_status()
    return cast(AzureOcrResponse, response.json())

def azure_read_result_to_djvu_xml(read_result: AzureOcrReadResult) -> str:
    paragraphs = []
    for block in read_result['blocks']:
        paragraph = etree.Element('PARAGRAPH')

        for line in block['lines']:
            line_el = etree.Element('LINE')
            for word in line['words']:
                word_el = etree.Element('WORD')
                rect = word['boundingPolygon']
                # coords is LBRT
                word_el.set('coords', ','.join(
                    str(x) for x in (
                        min(rect[0]["x"], rect[3]["x"]),
                        max(rect[2]["y"], rect[3]["y"]),
                        max(rect[1]["x"], rect[2]["x"]),
                        min(rect[0]["y"], rect[1]["y"]),
                    )
                ))
                word_el.set('x-confidence', str(100 * word['confidence']))
                word_el.text = word['text']
                line_el.append(word_el)
            paragraph.append(line_el)

        paragraphs.append(etree.tostring(paragraph, encoding='unicode'))
    return '\n'.join(paragraphs)

def ocr_djvu_page_azure(page_scan: PageScan) -> str:
    azure_response = fetch_ocr_azure(page_scan.image)

    return (
        f'<OBJECT type="image/x.djvu" width="{page_scan.width}" height="{page_scan.height}">\n'
        f'<PARAM name="DPI" value="{page_scan.dpi}"/>\n'
        '<HIDDENTEXT x-re-ocrd="true">'
        '<PAGECOLUMN>'
        '<REGION>'
        + azure_read_result_to_djvu_xml(azure_response['readResult']) +
        '</REGION>'
        '</PAGECOLUMN>'
        '</HIDDENTEXT>'
        '</OBJECT>'
    )




import importlib.resources
import json
import re
from typing import Literal, TypedDict, cast

import yaml

from tocky.extractor import TocEntry


class PromptSample(TypedDict):
    label: str | None
    input: str
    output: list[dict]  # But you know as a dict


# Read from a file
with importlib.resources.open_text("tocky", "prompt_samples.yaml") as f:
    PROMPT_SAMPLES = cast(dict[str, PromptSample], yaml.safe_load(f))

SAMPLE_FORMAT = """
{label}

Input:
```
{input}
```

Output:
```{output_format}
{output}
```
""".strip()

JSON_INSTRUCTIONS = """
The format you will need to output is JSON of this type:

```ts
Array<{
  level: number,
  label: string | null,
  title: string | null,
  pagenum: string | null,

  // Optional fields not present on most books
  authors: string[] | null,
  subtitle: string | null,
  description: string | null,
}>
```
""".strip()

MARKDOWN_INSTRUCTIONS = """
The format you will need to output is:

```
* {label (optional)} | {title} | {page number}
```
""".strip()

def format_toc(toc: list[TocEntry], extraction_format: Literal["json", "markdown"]) -> str:
    if extraction_format == "json":
        return '\n'.join([
            '[',
            '  ' + ',\n  '.join(json.dumps(entry.to_dict()) for entry in toc),
            ']',
        ])
    elif extraction_format == "markdown":
        return '\n'.join([
            chapter.to_markdown()
            for chapter in toc
        ])
    else:
        raise ValueError(f"Unknown extraction format: {extraction_format}")
   

def format_prompt_sample(sample: PromptSample, extraction_format: Literal["json", "markdown"]) -> str:
    input_str = sample["input"].strip('\n')
    input_str = re.sub(r'^\|', '', input_str, flags=re.MULTILINE)
    output_toc = [TocEntry(**entry) for entry in sample["output"]]
    return SAMPLE_FORMAT.format(
        label=sample.get("label", ""),
        input=input_str,
        output=format_toc(output_toc, extraction_format),
        output_format="" if extraction_format == "markdown" else extraction_format,
    ).strip()

def build_system_prompt(prompt_template: str, extraction_format: Literal["json", "markdown"]) -> str:
    format_instructions = JSON_INSTRUCTIONS if extraction_format == "json" else MARKDOWN_INSTRUCTIONS
    return prompt_template.format(
        format_instructions=format_instructions,
        PROMPT_SAMPLES={
            sample_id: format_prompt_sample(sample, extraction_format)
            for sample_id, sample in PROMPT_SAMPLES.items()
        }
    )

def process_extracted_output(output: str, extraction_format: Literal["json", "markdown"]) -> list[TocEntry]:
    toc_str = output.strip().strip('`')
    if extraction_format == 'json':
      if toc_str.startswith('json'):
        # Skip first line
        toc_str = '\n'.join(toc_str.split('\n')[1:])
      return [
        TocEntry(**entry)
        for entry in json.loads(toc_str)
      ]
    elif extraction_format == 'markdown':
      def clean_line(line: str) -> str:
        pipes = len(line.split('|')) - 1
        m = re.search(r'^ *\*+', line)
        if not m:
          return line
        stars = m.group()
        rest = line[m.span()[1]:]
        if pipes == 0:
          return f'{stars} | {rest.strip()} |'
        elif pipes == 1:
          return f'{stars} | {rest.strip()}'
        elif pipes == 2:
          return line
        elif pipes == 3:
          return line.replace('|', ' ', 1)
        else:
          # Not good
          return line

      return [
        TocEntry.from_markdown(clean_line(line))
        for line in toc_str.split('\n')
      ]
    else:
      raise ValueError(f"Unknown extraction format: {extraction_format}")



from typing import Literal
from lxml import etree
import re
from dataclasses import dataclass
import openai
import tiktoken

from tocky.extractor import AbstractExtractor, TocEntry, TocResponse
from tocky.extractor.formats import build_system_prompt, format_toc, process_extracted_output
from tocky.ocr.printer import print_ocr
from tocky.utils import avg_ocr_conf
from tocky.utils.ia import get_djvu_by_leaf_nums, get_ia_metadata, get_page_scan, ocaid_to_djvu_url


class BadOcrOnToc(Exception):
  pass


SYSTEM_PROMPT = """
You are a librarian extracting table of contents data in a structured, semantic format.

{format_instructions}

### Important Instructions:
- Fix any typos or bad characters generated by the OCR.
- Change any text in ALL UPPERCASE to be in Normal Title Case (e.g. "The Limit of a Function").
- Use the `label` field to store unimportant data like numerals (e.g. "Chapter 1", "Section 2.3").

### Examples:

{PROMPT_SAMPLES[basic]}

{PROMPT_SAMPLES[no_titles]}

{PROMPT_SAMPLES[nested]}
"""

PREV_TOC_PROMPT = """
The last table of contents was too long too fit in your context. Continue extracting the table of contents of "{book_title}".

Here is the end of what you last output. Do not output this again, though.

{prev_toc}

And here is the rest of the OCR text:
```
{ocr_text}
```
"""

TOC_PROMPT = """
Extract the table of contents from this OCR text of "{book_title}":

```
{ocr_text}
```
"""


@dataclass
class AiExtractorOptions:
  redo_ocr: bool = True
  ocr_engine: Literal['easyocr', 'tesseract', 'azure'] = 'easyocr'
  model: str = "gpt-4o-mini"
  max_sent_tokens: int = 1_000
  """
  GPT 4o mini can handle up to 128k input tokens and 16k output tokens.
  TODO: Experiment with larger input sizes.
  """
  extraction_format: Literal['json', 'markdown'] = 'json'


class AiExtractor(AbstractExtractor[AiExtractorOptions]):
  name = 'ai_extractor'

  def __init__(self):
    super().__init__()
    self.P = AiExtractorOptions()

  def extract(self, ocaid: str, detector_result: list[int]) -> list[TocEntry]:
    djvu_xml_to_fetch = set(detector_result) - set(self.S.ocr_cache.keys())
    if djvu_xml_to_fetch:
      djvu_url = ocaid_to_djvu_url(ocaid)
      start = min(djvu_xml_to_fetch)
      end = max(djvu_xml_to_fetch)
      for leaf_num, elem in get_djvu_by_leaf_nums(djvu_url, start, end):
        if leaf_num in djvu_xml_to_fetch:
          self.S.ocr_cache[leaf_num] = etree.tostring(elem, encoding='unicode')

    if self.P.redo_ocr:
      for leaf_num in detector_result:
        self.S.ocr_cache[leaf_num] = self.redo_ocr(ocaid, leaf_num, self.S.ocr_cache[leaf_num])

    self.toc_raw_ocr = [
      print_ocr(self.S.ocr_cache[leaf_num])
      for leaf_num in detector_result
    ]

    if re.search(r'([A-Za-z]{25,}|\beee+\b)', '\n'.join(self.toc_raw_ocr), flags=re.MULTILINE):
      raise BadOcrOnToc("Bad OCR on TOC")

    self.toc_response = self.extract_structured_toc(self.toc_raw_ocr, get_ia_metadata(ocaid)['metadata']['title'])

    return self.toc_response.toc

  def redo_ocr(self, ocaid: str, leaf_num: int, djvu_xml: str) -> str:
    from tocky.ocr import ocr_djvu_page

    root = etree.fromstring(djvu_xml)
    if root.xpath('.//HIDDENTEXT/@x-re-ocrd') == ['true']:
      return djvu_xml

    new_ocr = ocr_djvu_page(get_page_scan(ocaid, leaf_num), engine=self.P.ocr_engine)
    old_ocr_el = root.find('.//HIDDENTEXT')
    new_ocr_el = etree.fromstring(new_ocr).find('.//HIDDENTEXT')
    assert old_ocr_el and new_ocr_el
    if (avg_ocr_conf(new_ocr_el) or 100) > (avg_ocr_conf(old_ocr_el) or 0):
      root.replace(old_ocr_el, new_ocr_el)

    return etree.tostring(root, encoding='unicode')

  def extract_structured_toc(
    self,
    pages_ocr: list[str],
    book_title: str,
  ) -> TocResponse:
    structured_toc: list[TocEntry] = []
    prompt_tokens = 0
    completion_tokens = 0
    chunks = ['']
    for page_ocr in pages_ocr:
      extended_chunk = chunks[-1] + '\n' + page_ocr
      if len(tiktoken.encoding_for_model(self.P.model).encode(extended_chunk)) > self.P.max_sent_tokens:
        chunks.append(page_ocr)
      else:
        chunks[-1] += '\n' + page_ocr
    for chunk in chunks:
      toc_response = self.extract_structured_toc_chunk(chunk, book_title, prev_toc=structured_toc)
      structured_toc += toc_response.toc
      prompt_tokens += toc_response.prompt_tokens
      completion_tokens += toc_response.completion_tokens
    
    return TocResponse(structured_toc, prompt_tokens, completion_tokens)

  def extract_structured_toc_chunk(
    self,
    ocr_text: str,
    book_title: str,
    prev_toc: list[TocEntry] | None = None,
  ) -> TocResponse:
    if prev_toc:
      prev_toc_slice = prev_toc[-10:]
      if self.P.extraction_format == 'json':
        prev_toc_formatted = f'```json\n{format_toc(prev_toc_slice, self.P.extraction_format)}\n```'
      elif self.P.extraction_format == 'markdown':
        prev_toc_formatted = f'```\n{format_toc(prev_toc_slice, self.P.extraction_format)}\n```'
      else:
        raise ValueError(f"Unknown extraction format: {self.P.extraction_format}")

      message = PREV_TOC_PROMPT.format(
        book_title=book_title,
        prev_toc=prev_toc_formatted,
        ocr_text=ocr_text,
      )
    else:
      message = TOC_PROMPT.format(
        book_title=book_title,
        ocr_text=ocr_text,
      )
    print('openai request', message[0:500] + '...')

    system_prompt = build_system_prompt(SYSTEM_PROMPT, self.P.extraction_format)
    completion = openai.chat.completions.create(
      model=self.P.model,
      messages=[
        {"role": "system", "content": system_prompt},
        {"role": "user", "content": message},
      ],
      # max_tokens=1024,
      n=1,
      stop=None,
      temperature=0.5,
    )

    assert completion.choices[0].message.content
    assert completion.usage

    toc = process_extracted_output(
      completion.choices[0].message.content,
      self.P.extraction_format,
    )
    return TocResponse(
        toc,
        completion.usage.prompt_tokens,
        completion.usage.completion_tokens,
    )



from dataclasses import dataclass
from typing import Generic, TypeVar

from tocky.utils import ShareableState

@dataclass
class TocEntry:
    level: int
    label: str | None = None
    title: str | None = None
    pagenum: str | None = None
    
    authors: list[str] | None = None
    subtitle: str | None = None
    description: str | None = None

    @staticmethod
    def from_markdown(line: str) -> 'TocEntry':
        """
        >>> TocEntry.from_markdown("* 1 | Introduction | 1")
        TocEntry(level=1, label='1', title='Introduction', pagenum='1')
        >>> TocEntry.from_markdown("** 1.1 | Subsection | 2")
        TocEntry(level=2, label='1.1', title='Subsection', pagenum='2')
        >>> TocEntry.from_markdown("| Chapter 1 | 3")
        TocEntry(level=0, label=None, title='Chapter 1', pagenum='3')
        """
        parts = line.split('|')
        if len(parts) != 4:
            raise ValueError(f"Invalid line: {line}")
        
        def count_stars(s: str) -> int:
            i = 0
            for i, c in enumerate(s):
                if c != '*':
                    return i
            return i

        prefix, title, pagenum = parts
        return TocEntry(
            level=count_stars(prefix.lstrip()) - 1,
            label=prefix.lstrip('*').strip() or None,
            title=title.strip() or None,
            pagenum=pagenum.strip() or None,
        )

    def to_markdown(self) -> str:
        """
        >>> TocEntry(level=1, label='1', title='Introduction', pagenum='1').to_markdown()
        '    ** 1 | Introduction | 1'
        >>> TocEntry(level=2, label='1.1', title='Subsection', pagenum='2').to_markdown()
        '        *** 1.1 | Subsection | 2'
        >>> TocEntry(level=0, label=None, title='Chapter 1', pagenum='3').to_markdown()
        '* | Chapter 1 | 3'
        >>> TocEntry(level=0, label=None, title='Chapter 1', pagenum=None).to_markdown()
        '* | Chapter 1 | '
        >>> TocEntry(level=0, label=None, title=None, pagenum=None).to_markdown()
        '* |  | '
        """
        # Error if any other field is not None
        if self.authors or self.subtitle or self.description:
            raise ValueError("Cannot convert to markdown with authors, subtitle, or description")

        result = ('    ' * self.level) + ('*' * (self.level + 1))
        if self.label:
            result += f' {self.label}'
        result += f" | {self.title} | {self.pagenum or ''}"
        return result.rstrip()

    def to_dict(self) -> dict:
        """
        >>> TocEntry(level=1, label='1', title='Introduction', pagenum='1').to_dict()
        {'level': 1, 'label': '1', 'title': 'Introduction', 'pagenum': '1'}
        """
        # Remove all None
        return {k: v for k, v in self.__dict__.items() if v is not None}


@dataclass
class TocResponse:
  toc: list[TocEntry]
  prompt_tokens: int
  completion_tokens: int


TParams = TypeVar("TParams")

class AbstractExtractor(Generic[TParams]):
    name: str
    P: TParams
    S: ShareableState
    debug = True
    """
    When debug is set to true, extra helper variables could be set
    """

    def __init__(self):
        self.S = ShareableState()

    toc_raw_ocr: list[str] | None = None
    toc_response: TocResponse | None = None

    def predict_cost(self) -> float:
        raise NotImplementedError()

    def extract(self, ocaid: str, detector_result: list[int]) -> list[TocEntry]:
        raise NotImplementedError()



from dataclasses import dataclass
import itertools
import re
from typing import Literal

from lxml import etree

from tocky.detector import AbstractDetector
from tocky.utils.ia import extract_leaf_num, get_djvu_pages, get_page_scan, ocaid_to_djvu_url
from tocky.ocr import ocr_djvu_page
from tocky.utils import avg_ocr_conf

TOC_PAGE_DETECTOR_VERSION = [
    ('v2.E.4', 'Check more pages for TOC if TOC has begun'),
    ('v2.E.3', 'Handle pages with all non-numeric pagenums'),
    ('v2.E.2', 'Add support for appendix page numbers like [A-C]123'),
    ('v2.E.1', 'Fix: Don\'t check every page after TOC found!'),
    ('v2.E.0', 'Add support for re-ocring when ending TOC range to avoid missing last page(s)'),
    ('v2.D.0', 'Add support for re-ocring previous page once first TOC page detected to avoid missing first page'),
]

POSSIBLE_PAGENUM_RE = re.compile(r'\b([ABC]?\d+|[xvil]+)$', re.IGNORECASE)
NEGATIVE_RE = re.compile(r'\d[.,-]\d+$')

def match_possible_pagenum(text: str) -> re.Match | None:
  """
  Matches things like:
  - 1
  - A1 (eg https://archive.org/details/calculusearlytra0000stew_08th/page/n13)
  - xii

  >>> bool(match_possible_pagenum('1'))
  True
  >>> bool(match_possible_pagenum('Appendix 1: Science A1'))
  True
  >>> bool(match_possible_pagenum('preface xii'))
  True
  >>> bool(match_possible_pagenum('1.32'))
  False
  >>> bool(match_possible_pagenum('1,32'))
  False
  >>> bool(match_possible_pagenum('1-32'))
  False
  """
  if (m := POSSIBLE_PAGENUM_RE.search(text)) and NEGATIVE_RE.search(text) is None:
    return m
  else:
    return None

def pagenum_word_to_int(n: str) -> int | None:
  """
  >>> pagenum_word_to_int('1')
  1
  >>> pagenum_word_to_int('A1')
  1
  >>> pagenum_word_to_int('xii')
  >>> pagenum_word_to_int('   432')
  432
  """
  n = n.strip()
  if n.isnumeric():
    return int(n)
  elif n[0].isalpha() and n[1:].isnumeric():
    return int(n[1:])
  else:
    return None

@dataclass
class OcrDetectorOptions:
  ocr_engine: Literal['easyocr', 'tesseract', 'azure'] = 'easyocr'
  allow_reocr: bool = True

class OcrDetector(AbstractDetector[OcrDetectorOptions]):
  name = 'ocr_detector'

  def __init__(self):
    super().__init__()
    self.P = OcrDetectorOptions()

  def predict_cost(self):
    return 0

  def detect(self, ocaid: str) -> list[int]:
    results = list(self.extract_toc_pages(ocaid))
    self.S.ocr_cache.update({
        extract_leaf_num(page_name): djvu_xml_str
        for (page_name, djvu_xml_str) in results
    })
    return [extract_leaf_num(page_name) for (page_name, _) in results]
  
  def extract_toc_pages(self, ocaid: str):
    return self.detect_table_of_contents_pages(ocaid_to_djvu_url(ocaid))

  def detect_table_of_contents_pages(self, djvu_url: str):
    has_begun = False
    for (page_name, elem_str, toc_analysis) in self.analyze_djvu_for_toc(djvu_url):
      if has_begun and not toc_analysis.is_toc:
        break
      if toc_analysis.is_toc:
        has_begun = True
        yield (page_name, elem_str)

  def analyze_djvu_for_toc(self, djvu_url: str):
    P = TOCDetectorHyperParams()
    has_begun = False
    reocrs_done = 0
    last_result: tuple[str, etree._Element, PageTocDetectionResult] = None
    for page_name, elem in get_djvu_pages(djvu_url, 1, end=None):
      toc_analysis = self.analyze_page_for_toc(elem, has_begun, allow_reocr=self.P.allow_reocr and (has_begun or reocrs_done < 10))
      if toc_analysis.reran_ocr:
        reocrs_done += 1
      # Let's check predecessor with redone ocr
      if self.P.allow_reocr and not has_begun and toc_analysis.is_toc and last_result and not last_result[2].reran_ocr:
        # Check if number is too high before redoing OCR
        recheck_previous = False
        for line in elem.findall(".//LINE"):
            words = line.findall(".//WORD")
            if words:
              last_word = words[-1]
              # Roman numerals are usually at the beginning of the book,
              # so need to check the previous page
              if bool(re.search(r'\b([xvil]+)$', last_word.text)):
                recheck_previous = False
                break

              # If it's just a normal numbers, if the number is too high,
              # then we should recheck the previous page
              if number_m := re.search(r'\b(\d+)$', last_word.text):
                n = int(number_m[1])
                recheck_previous = n > 25
                break
            
              # Other page numbers, like A1, B1, C1, etc are usually at
              # the end of the book in the appendix, so we should recheck
              # the previous page
              if match_possible_pagenum(last_word.text):
                recheck_previous = True
                break

        if recheck_previous:
          new_toc_analysis = self.analyze_page_for_toc(last_result[1], False, redo_ocr=True)
          reocrs_done += 1
          if new_toc_analysis.is_toc:
            last_result = (last_result[0], last_result[1], new_toc_analysis)

      if self.P.allow_reocr and has_begun and not toc_analysis.is_toc and not toc_analysis.reran_ocr:
        # Let's try again with rerun ocr
        reocrs_done += 1
        toc_analysis = self.analyze_page_for_toc(elem, has_begun, redo_ocr=True)

      # Yield the thing we know won't change anymore
      if last_result:
        yield (last_result[0], etree.tostring(last_result[1], encoding='unicode'), last_result[2])

      has_begun = toc_analysis.is_toc
      last_result = (page_name, elem, toc_analysis)

      leaf_num = extract_leaf_num(page_name)
      if has_begun and leaf_num > P.max_leaf_to_check_while_in_toc:
        break
      if not has_begun and leaf_num > P.max_leaf_to_check:
        break

    if last_result:
      yield (last_result[0], etree.tostring(last_result[1], encoding='unicode'), last_result[2])

  def analyze_page_for_toc(
    self,
    elem: etree._Element,
    has_begun = False,
    allow_reocr=True,
    redo_ocr=False,
  ) -> 'PageTocDetectionResult':
    hiddentext = elem.find(".//HIDDENTEXT")
    if hiddentext is None:
      return PageTocDetectionResult(False, 'Missing hiddentext')

    P = TOCDetectorHyperParams()
    result = PageTocDetectionResult()

    page_file: str = elem.xpath('./@usemap')[0]
    ocaid = page_file.rsplit('_', 1)[0]
    leaf_num = extract_leaf_num(page_file)
    # print('analyze_page_for_toc', page_file, f'{allow_reocr=}')

    def do_redo_ocr(min_conf: float):
      if elem.xpath('.//HIDDENTEXT/@x-re-ocrd') == ['true']:
        return
      new_ocr = ocr_djvu_page(get_page_scan(ocaid, leaf_num), engine=self.P.ocr_engine)
      result.reran_ocr = True
      new_ocr_el = etree.fromstring(new_ocr).find('.//HIDDENTEXT')
      # Re-OCR will always output confidences, so assume 100 (likely empty page)
      new_conf = avg_ocr_conf(new_ocr_el) or 100
      # print(f'{page_file} mean_conf {(mean_conf or 0):.2f} -> {new_conf:.2f}')
      if new_conf > min_conf:
        result.used_new_ocr = True
        elem.replace(elem.find('.//HIDDENTEXT'), new_ocr_el)
        return self.analyze_page_for_toc(elem, has_begun, allow_reocr=False, redo_ocr=False)

    mean_conf = avg_ocr_conf(elem)

    if redo_ocr:
      if new_result := do_redo_ocr(min_conf=(mean_conf or 0) - 15):
        return new_result

    if mean_conf is not None and mean_conf < P.min_mean_word_confidence:
      if allow_reocr:
        if new_result := do_redo_ocr(min_conf=mean_conf):
          return new_result
      return PageTocDetectionResult(False, f'Mean page confidence too low ({mean_conf:.2f} < {P.min_mean_word_confidence})')

    # If there are more than 40 words and the average right coord of each word is
    # less than 2/3 width of page, then potentially should re-ocr cause mis-OCR'd
    # pages
    if allow_reocr:
      word_rights = [
          int(coord.split(',')[2])
          for coord in hiddentext.xpath('.//WORD/@coords')
      ]
      if len(word_rights) > 70:
        width = int(elem.xpath('./@width')[0])
        words_near_right = len([r for r in word_rights if r > width * .8 ])

        words_near_right_text = [
            word.text
            for word in hiddentext.xpath('.//WORD')
            if int(word.xpath('./@coords')[0].split(',')[2]) > width * .8
        ]
        numeric_words_near_right = len([
            word
            for word in words_near_right_text
            if match_possible_pagenum(word)
        ])
        # print(f'{words_near_right=}, {numeric_words_near_right=} of {len(word_rights)=}')
        if (words_near_right - numeric_words_near_right) < 3:
          if new_result := do_redo_ocr(min_conf=(mean_conf or 0) - 15):
            return new_result


    numbers_at_end_of_lines = []
    for line in hiddentext.findall(".//LINE"):
        words = line.findall(".//WORD")
        if words:
          last_word = words[-1]
          if m := match_possible_pagenum(last_word.text):
              confidence = last_word.xpath('./@x-confidence')
              bad_ocr = confidence and float(confidence[0]) < P.min_word_confidence
              if len(m.group()) == 1 and bad_ocr:
                # Exclude, likely noise
                pass
              else:
                numbers_at_end_of_lines.append(m.group())

    from collections import Counter
    max_repeated = max(Counter(numbers_at_end_of_lines).values(), default=0)
    lines_w_uniq_nums = len(set(numbers_at_end_of_lines))

    if max_repeated > P.max_repeated_nums:
      return PageTocDetectionResult(False, 'Too many repeated numbers')

    if lines_w_uniq_nums < P.min_lines_with_nums:
      return PageTocDetectionResult(False, f'Not enough lines with nums ({lines_w_uniq_nums} < {P.min_lines_with_nums})')

    if lines_w_uniq_nums > P.max_lines_with_nums:
      return PageTocDetectionResult(False, 'Too many lines with nums')

    nums = [
        num
        for n in numbers_at_end_of_lines
        if (num := pagenum_word_to_int(n)) is not None
    ]

    if len(nums) > 4:
      increasing_count = 0
      non_increasing_count = 0
      for (prev, cur) in itertools.pairwise([0] + nums):
        if prev <= cur:
          increasing_count += 1
        else:
          non_increasing_count += 1

      increasing_percent = increasing_count / (increasing_count + non_increasing_count)
      if increasing_percent < P.min_increasing_percent:
        return PageTocDetectionResult(False, f'Not enough of the nums are increasing {increasing_percent:.2f} < {P.min_increasing_percent}')

    if has_begun:
      first_line = ' '.join(word.text for word in hiddentext.find(".//LINE").findall(".//WORD"))
      if 'illustration' in first_line.lower() or 'index' in first_line.lower() or 'figure' in first_line.lower() or 'tables' in first_line.lower():
        return PageTocDetectionResult(False, 'Contains illustration/index')
    else:
      small_nums_count = len([n for n in numbers_at_end_of_lines if len(n) < 4])
      if small_nums_count < P.min_small_nums_count_for_first:
        return PageTocDetectionResult(False, 'Not enough small numbers for first TOC page')

      # Check average word length probably bad OCR on empty page
      word_lens = [
          len(word.text.strip())
          for word in elem.findall('.//WORD')
          # Skip non-word-y things
          if not match_possible_pagenum(word.text) and not re.search(r'^[\.^;:{}]+$', word.text.strip())
      ]
      avg_word_len = sum(word_lens) / len(word_lens)
      if avg_word_len < P.min_avg_word_len:
        return PageTocDetectionResult(False, f'Average word length too low {avg_word_len:.2f} < {P.min_avg_word_len}')

    if allow_reocr:
      if new_result := do_redo_ocr(min_conf=(mean_conf or 0) - 15):
        return new_result

    return PageTocDetectionResult(True)


@dataclass
class TOCDetectorHyperParams:
  min_lines_with_nums: int = 5
  max_lines_with_nums: int = 50
  max_repeated_nums: int = 4
  min_word_confidence: float = 40
  min_mean_word_confidence: float = 30
  min_small_nums_count_for_first: int = 4
  min_avg_word_len: float = 3.8
  min_increasing_percent: float = 0.75
  max_leaf_to_check: int = 28
  max_leaf_to_check_while_in_toc: int = 30


@dataclass
class PageTocDetectionResult:
  is_toc: bool = False
  failure: str | None = None
  reran_ocr: bool = False
  used_new_ocr: bool = False



from collections.abc import Iterable, Iterator
import functools
from io import BytesIO
import itertools
import json
import re
import sys
from typing import Literal, TypedDict, cast
from internetarchive import get_session
from PIL import Image
from lxml import etree
import pycountry
from requests import HTTPError
import concurrent.futures
from dotenv import load_dotenv
import requests

from tocky.utils import PageScan

load_dotenv()  # take environment variables from .env.

ia_session = get_session()

class IaMetadata(TypedDict):
  identifier: str
  ppi: str
  language: str | list[str]
  """Usually 3 letter MARC code"""
  title: str
  imagecount: str

class IaFileMetadata(TypedDict):
  name: str
  source: Literal['derivative', 'original', 'metadata']
  format: str
  mtime: str
  size: str
  md5: str
  filecount: str


class IaFullMetadata(TypedDict):
  metadata: IaMetadata
  files: list[IaFileMetadata]


@functools.cache
def get_ia_metadata(ocaid: str) -> IaFullMetadata:
  return ia_session.get(f"https://archive.org/metadata/{ocaid}").json()

def get_page_image(
  ocaid: str,
  leaf_num: int,
  ext='jpg',
  reduce=2,
  quality=80,
  jp2_zip=None,
  stream=False,
) -> Image.Image:
  jp2_zip = jp2_zip or f"{ocaid}_jp2.zip"
  file_prefix = jp2_zip.replace('_jp2.zip', '')
  url = f"https://archive.org/download/{ocaid}/{jp2_zip}/{jp2_zip.replace('.zip', '')}%2F{file_prefix}_{leaf_num:04}.jp2"
  img = ia_session.get(
    url,
    params={
      'ext': ext,
      'reduce': str(reduce),
      'quality': str(quality),
    },
    stream=stream
  )

  if stream:
    return img  # TODO: Fix type

  img.raise_for_status()

  return Image.open(BytesIO(img.content))

def get_page_scan(
  ocaid: str,
  leaf_num: int,
  ext = 'jpg',
  reduce = 0,
  quality = 100,
  image: Image.Image | None = None
) -> PageScan:
  """
  :param ocaid: The identifier of the item on archive.org
  :param leaf_num: The leaf number of the page to fetch
  :param image: An optional PIL Image object to use instead of fetching the image from the internet
  """
  full_metadata = get_ia_metadata(ocaid)

  if not image:
    jp2_zip = get_main_jp2_zip(full_metadata)
    image = get_page_image(ocaid, leaf_num, ext=ext, reduce=reduce, quality=quality, jp2_zip=jp2_zip)

  md_lang = full_metadata['metadata']['language']
  lang = md_lang if isinstance(md_lang, str) else md_lang[0]

  return PageScan(
    uri=f'https://archive.org/details/{ocaid}#page/leaf{leaf_num}',
    image=image,
    dpi=int(full_metadata['metadata']['ppi']),
    lang=ia_language_to_iso639_2_code(lang) or 'eng',
  )


def get_main_jp2_zip(ocaid_or_metadata: str | IaFullMetadata) -> str:
  if isinstance(ocaid_or_metadata, str):
    full_metadata = get_ia_metadata(ocaid_or_metadata)
  else:
    full_metadata = ocaid_or_metadata

  all_jp2_zips = list(
    file
    for file in full_metadata['files']
    if file['name'].endswith('_jp2.zip') and not file['name'].endswith('_raw_jp2.zip')
  )
  assert all_jp2_zips, "No jp2 files found"
  assert len(all_jp2_zips) == 1, f"Multiple jp2 files found {[f['name'] for f in all_jp2_zips]}"
  jp2_zip = all_jp2_zips[0]
  return jp2_zip['name']


def get_book_images(
  ocaid: str,
  leaf_nums: Iterable[int],
  reduce = 3,
) -> Iterator[Image.Image]:
  jp2_zip = get_main_jp2_zip(ocaid)
  def get_image(leaf_num: int):
    # try:
      return get_page_image(ocaid, leaf_num, reduce=reduce, jp2_zip=jp2_zip)
    # except HTTPError as e:
    #   # Return an empty image of size 100x300
    #   return Image.new('RGB', (100, 300))
    
  with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
    return executor.map(get_image, leaf_nums)

def ia_language_to_iso639_2_code(lang: str) -> str | None:
  if len(lang) == 3:
    return lang
  else:
    language = pycountry.languages.get(name=lang)
    if language:
      return language.alpha_3
    else:
      return None

def extract_leaf_num(page_filename: str) -> int:
  """
  E.g. 'goody_0001.djvu' -> 1
  """
  m = re.search(r'(?:_)(\d+)(?:\.djvu)', page_filename)
  if not m:
    raise ValueError(f"Unable to extract leaf number from {page_filename}")
  return int(m.group(1))

def ocaid_to_djvu_url(ocaid: str) -> str:
  get_ia_metadata(ocaid)
  return f"https://archive.org/download/{ocaid}/{ocaid}_djvu.xml"

def get_djvu_pages(djvu_url: str, start: int=0, end: int=sys.maxsize):
    response = ia_session.get(djvu_url, stream=True)

    if response.status_code != 200:
        print("Error: Unable to fetch the DJVU file.")
        return

    response.raw.decode_content = True
    for _, elem in itertools.islice(etree.iterparse(response.raw, events=("end",), tag="OBJECT"), start, end):
      page_name = cast(str, elem.xpath(".//PARAM[@name='PAGE']/@value")[0])
      yield page_name, cast(etree._Element, elem)

def get_djvu_by_leaf_nums(djvu_url: str, start: int, end: int):
  """
  Fetches the Djvu XML for a range of leaf numbers, inclusive
  """

  for page_name, elem in get_djvu_pages(djvu_url):
    leaf_num = extract_leaf_num(page_name)
    if start <= leaf_num <= end:
      yield leaf_num, elem
    if leaf_num > end:
      break

class IaLiteMetadata(TypedDict):
  identifier: str
  openlibrary_edition: str | None

def bulk_ia_to_ol(ia_records: list[IaLiteMetadata]) -> dict[str, dict]:
  ol_records_by_key = requests.get('http://openlibrary.org/api/get_many', params={
    'keys': json.dumps([
      f'/books/{metadata["openlibrary_edition"]}'
      for metadata in ia_records
    ])
  }).json()['result']
    
  for ia_record in ia_records:
    ol_record = ol_records_by_key[f'/books/{ia_record["openlibrary_edition"]}']
    if 'ocaid' not in ol_record:
      ol_record['ocaid'] = ia_record['identifier']
  
  return ol_records_by_key



from collections.abc import Callable
from dataclasses import dataclass, field
import functools
import re
from time import time
from traceback import TracebackException
from typing import Generic, Literal, TypeVar, cast
from lxml import etree
from PIL import Image


@dataclass
class ShareableState:
  """
  Shareable state between detector runs or between detectors/extractors
  """
  ocr_cache: dict[int, str] = field(default_factory=dict)
  """Map of leaf number to OCR text (djvu xml format)"""


T = TypeVar('T')


@dataclass
class ResultStatSuccess(Generic[T]):
    success: Literal[True]
    time: float
    result: T
    error: None = None
    traceback: None = None

    def to_dict(self):
        return {
            'success': self.success,
            'time': self.time,
            'result': self.result,
        }

@dataclass
class ResultStatError:
    success: Literal[False]
    time: float
    result: None
    error: Exception
    
    @property
    def traceback(self) -> str:
        return '\n'.join(TracebackException.from_exception(self.error).format())

    def to_dict(self):
        return {
            'success': self.success,
            'time': self.time,
            'result': None,
            'error': str(self.error),
            'traceback': self.traceback,
        }
ResultStat = ResultStatSuccess[T] | ResultStatError

def run_with_result_stats(func: Callable[[], T]) -> ResultStat[T]:
    start = time()
    result = None
    try:
        result = func()
        error = None
    except Exception as e:
        error = e
    finally:
        end = time()
        dur = end - start
        if error is None:
            return ResultStatSuccess(
                success=True,
                time=dur,
                result=cast(T, result),
                error=None,
                traceback=None,
            )
        else:
            return ResultStatError(
                success=False,
                time=dur,
                result=None,
                error=error,
            )


@dataclass
class PageScan:
  uri: str
  image: Image.Image
  dpi: int
  lang: str

  @property
  def width(self) -> int:
    return self.image.width
  
  @property
  def height(self) -> int:
    return self.image.height

  def preview(self, width=300) -> Image.Image:
    return self.image.resize((width, int(width * self.image.height / self.image.width)))


@dataclass
class Rect:
  x: int
  y: int
  width: int
  height: int
  left: int
  right: int
  bottom: int
  top: int

  @staticmethod
  def from_xywh(rect: tuple) -> 'Rect':
    x, y, w, h = rect
    return Rect(
        x=x,
        y=y,
        width=w,
        height=h,
        left=x,
        right=x + w,
        bottom=y + h,
        top=y,
    )

  @staticmethod
  def from_ltrb(rect: tuple) -> 'Rect':
    l, t, r, b = rect
    return Rect(
        x=l,
        y=t,
        width=r - l,
        height=b - t,
        left=l,
        right=r,
        bottom=b,
        top=t,
    )

  @staticmethod
  def from_cw_points(points: list[list[int]]) -> 'Rect':
    # Points can be rotated
    l = min(p[0] for p in points)
    t = min(p[1] for p in points)
    r = max(p[0] for p in points)
    b = max(p[1] for p in points)
    return Rect.from_ltrb((l, t, r, b))

  def to_ltrb(self) -> tuple:
    return (self.left, self.top, self.right, self.bottom)

def pretty_print_xml(root: etree._Element | str) -> str:
  if isinstance(root, str):
    root = etree.fromstring(root)

  import xml.dom.minidom
  result = xml.dom.minidom.parseString(etree.tostring(root)).toprettyxml(indent='  ')
  return strip_newlines(result)

def strip_newlines(text: str) -> str:
  result = re.sub(r'^\s+$', '', text, flags=re.MULTILINE)
  result = re.sub(r'\n+', '\n', result)
  return result

def avg_ocr_conf(el: etree._Element) -> float:
  conf_sum = 0
  conf_count = 0
  for word in el.findall('.//WORD'):
    if word.xpath('./@x-confidence'):
      conf_sum += float(word.xpath('./@x-confidence')[0])
      conf_count += 1
  if conf_count:
    return conf_sum / conf_count


@functools.cache
def get_git_sha() -> str | None:
  import subprocess

  try:
    return subprocess.check_output(['git', 'rev-parse', 'HEAD']).strip().decode('utf-8')
  except subprocess.CalledProcessError:
    return None

@functools.cache
def get_tocky_version() -> str:
  from importlib.metadata import version, PackageNotFoundError
  try:
    tocky_version = version('tocky')
  except PackageNotFoundError:
    tocky_version = 'local'

  if git_sha := get_git_sha():
    return f'{tocky_version}+{git_sha[:8]}'
  else:
    return tocky_version



import re

from tocky.extractor import TocEntry

class OlTocParseError(Exception):
  pass


def parse_ol_toc(toc_string: str) -> list[TocEntry]:
    toc_entries = []
    lines = toc_string.strip().split('\n')
    pattern = r'''
        ^\s*
        (?P<stars>\*+)      # Capture stars
        \s*
        \|
        \s*
        (?P<label>[^|]+)     # Capture label
        \s*
        \|
        \s*
        (?P<pagenum>\S.*)     # Capture page number
        $                    # End of line
    '''

    for line in lines:
        # Chat GPT started outputting these lately
        if line.strip() == '```':
          continue
        parts = line.strip().split('|')
        if len(parts) != 3:
          raise OlTocParseError()
        label, title, pagenum = parts
        if not label.startswith('*'):
          raise OlTocParseError()
        toc_entries.append(
            TocEntry(
              level=len(re.search(r'\*+', label)[0]),
              label=label.lstrip('* ').rstrip(),
              title=title.strip(),
              pagenum = pagenum.strip(),
          )
        )

    return toc_entries


def validate_extracted_toc(parsed_toc: list[TocEntry], num_pages: int):
  for entry in parsed_toc:
    if len(entry.label or '') > 160 or len(entry.title or '') > 160:
      return 'Label/Title Too Long'

  nums = [
    int(entry.pagenum)
    for entry in parsed_toc
    if entry.pagenum and entry.pagenum.isnumeric()
  ]
  roman_numerals = [
    entry.pagenum
    for entry in parsed_toc
    if entry.pagenum and re.search(r'^[xvi]+$', entry.pagenum, flags=re.IGNORECASE)
  ]

  if not (nums + roman_numerals):
    return 'No Numbers'

  if len(nums + roman_numerals) < 2:
    return 'Not Enough Numbers'

  if len(nums + roman_numerals) < (len(parsed_toc) * 0.7):
    return 'Too Few Lines With Numbers'

  if not roman_numerals and (nums[0] > 40 and nums[1] > 40):
    return 'Starts Too High'

  if num_pages > 150 and nums[-1] < (num_pages - 150) and nums[-2] < (num_pages - 150):
    return 'End Too Low'

  return 'Valid'


import base64
from dataclasses import dataclass
import json
import re
import textwrap
from io import BytesIO

from openai import OpenAI
from PIL import Image, ImageDraw, ImageFont

from tocky.detector import AbstractDetector
from tocky.utils.ia import get_book_images
from tocky.utils.llm import MODEL_PRICES

SYSTEM_PROMPT: str = """
You are a bot that helps in the detection of all table of contents pages in a book.

Notes:
- Make sure to get all the pages, not just the first page of the table of contents
- AVOID things like the copyright page or table of figures/illustrations
- If you cannot detect a table of contents, output an empty array instead of guessing

Please output only JSON of this format: { "toc_pages": [7,8], "notes": "<anything you want to share>" }
"""

@dataclass
class AiVisionDetectorOptions:
    model: str = "gpt-4o-mini"
    max_tokens: int = 200
    image_size: tuple[int, int] = (1024, 512)

class AiVisionDetector(AbstractDetector[AiVisionDetectorOptions]):
    """
    This detector uses multi-modal AI models ability to process images and text together
    to detect table of contents pages in a book.
    """
    name = 'ai_vision_detector'

    def __init__(self):
        super().__init__()
        self.P = AiVisionDetectorOptions()

    @property
    def model(self):
        return MODEL_PRICES[self.P.model]

    def predict_cost(self):
        return self.model.predict_cost([SYSTEM_PROMPT], self.P.max_tokens, [self.P.image_size])

    def detect(self, ocaid: str):
        small_images = list(get_book_images(ocaid, range(0, 28), reduce=3))
        composite_image = place_images_in_grid(
            small_images,
            composite_width=self.P.image_size[0],
            composite_height=self.P.image_size[1],
            image_width=140,
        )

        if self.debug:
            self.small_images = small_images
            self.composite_image = composite_image

        client = OpenAI()
        response = client.chat.completions.create(
            model=self.P.model,
            messages=[
                {
                    "role": "system",
                    "content": SYSTEM_PROMPT,
                },
                {
                    "role": "user",
                    "content": [
                        {
                            "type": "image_url",
                            "image_url": {
                                "url": f"data:image/png;base64,{image_to_base64(composite_image)}",
                                # "detail": "low",
                            }
                        },
                    ],
                }
            ],
            max_tokens=self.P.max_tokens,
        )

        if self.debug:
            self.response = response

        response_str = response.choices[0].message.content
        assert response_str
        json_m = re.search(r'\{[\s\S]*\}', response_str, flags=re.MULTILINE)
        assert json_m

        return json.loads(json_m.group(0))['toc_pages']



def place_images_in_grid(images, composite_width=1024, composite_height=512, image_width=100):
    # Set the larger image height as twice the desired composite height
    larger_image_height = composite_height * 2

    # Calculate the number of images per row and initial image height
    num_images_per_row = composite_width // image_width
    spacing_x = (composite_width % image_width) // (num_images_per_row - 1) if num_images_per_row > 1 else 0
    image_height = min((img.size[1] * image_width) // img.size[0] for img in images)

    # Create a larger composite image with a white background
    larger_composite_image = Image.new('RGB', (composite_width, larger_image_height), (0,0,0))

    # You may need to specify a font file if the default is not acceptable
    font = ImageFont.load_default()

    # Starting position
    x_offset, y_offset = 0, 0

    # Place images in the larger composite image
    for i, img in enumerate(images):
        img_resized = img.resize((image_width, image_height), Image.ANTIALIAS)

        # Create a drawing context to draw on the resized image
        draw_img = ImageDraw.Draw(img_resized)

        # Define the text to draw
        text = str(i)
        text_width, text_height = draw_img.textsize(text, font=font)

        # Calculate position for text, rectangle size and draw a white rectangle behind the text
        text_x, text_y = 5, 5
        rect_x1, rect_y1 = text_x - 2, text_y - 2
        rect_x2, rect_y2 = text_x + text_width + 2, text_y + text_height + 2
        draw_img.rectangle((rect_x1, rect_y1, rect_x2, rect_y2), fill="white")

        # Draw the text over the rectangle
        draw_img.text((text_x, text_y), text, font=font, fill="black")

        # Check if the next image fits in the current row and adjust offsets appropriately
        if x_offset + image_width > composite_width:
            x_offset = 0
            y_offset += image_height + 10  # Add 10 px vertical spacing

        # Paste the image with text into the larger composite
        larger_composite_image.paste(img_resized, (x_offset, y_offset))

        # Update the x_offset for the next image
        x_offset += image_width + spacing_x

    # Finally, crop the larger composite image to the desired final size
    final_composite_image = larger_composite_image.crop((0, 0, composite_width, composite_height))

    return final_composite_image


def image_to_base64(pil_image, format="PNG"):
    # Create an in-memory bytes buffer
    buffer = BytesIO()
    # Save the image to the buffer in the specified format
    pil_image.save(buffer, format=format)
    # Get the raw bytes of the image data from the buffer
    img_bytes = buffer.getvalue()
    # Encode the bytes to base64
    img_base64 = base64.b64encode(img_bytes)
    # Convert bytes to string for easier usage
    img_base64_str = img_base64.decode('utf-8')
    return img_base64_str

def concatenate_and_resize(images, target_height=512):
    # Get the total width of all images combined
    total_width = sum(img.size[0] for img in images)

    # Calculate the height of the tallest image (to set the row height)
    max_height = max(img.size[1] for img in images)

    scale_factor = target_height / max_height
    target_width = int(total_width * scale_factor)

    combined_image = Image.new('RGB', (target_width, target_height))

    # Concatenate images into one big row
    x_offset = 0
    for img in images:
        new_width = int(img.size[0] * scale_factor)
        resized_img = img.resize((new_width, target_height), Image.ANTIALIAS)

        # Paste the resized image into the combined image
        combined_image.paste(resized_img, (x_offset, 0))
        x_offset += new_width

    return combined_image



from dataclasses import dataclass
import functools
import os
from pathlib import Path


@functools.cache
def get_env() -> 'TockyEnv':
    return TockyEnv.from_env()


@dataclass
class TockyEnv:
    TOCKY_SERVER_NAME: str
    """e.g. 'localhost:5000'"""

    TOCKY_APPLICATION_ROOT: str
    """e.g. '/tocky' or ''"""

    TOCKY_PREFERRED_URL_SCHEME: str
    """e.g. 'http' or 'https'"""

    TOCKY_QUEUE_DB_PATH: Path
    
    @property
    def TOCKY_SERVER_KEY(self) -> str:
        return os.environ['TOCKY_SERVER_KEY']
    
    @property
    def TOCKY_USER_KEY(self) -> str:
        return os.environ['TOCKY_USER_KEY']
    
    @property
    def AZURE_SUBSCRIPTION_KEY(self) -> str | None:
        return os.environ.get('AZURE_SUBSCRIPTION_KEY')
    
    @property
    def AZURE_ENDPOINT(self) -> str | None:
        return os.environ.get('AZURE_ENDPOINT')

    def get_app_prefix(self) -> str:
        return f'{self.TOCKY_PREFERRED_URL_SCHEME}://{self.TOCKY_SERVER_NAME}{self.TOCKY_APPLICATION_ROOT}'

    @staticmethod
    def from_env() -> 'TockyEnv':
        return TockyEnv(
            TOCKY_SERVER_NAME=os.environ['TOCKY_SERVER_NAME'].rstrip('/'),
            TOCKY_APPLICATION_ROOT=os.environ['TOCKY_APPLICATION_ROOT'].rstrip('/'),
            TOCKY_PREFERRED_URL_SCHEME=os.environ['TOCKY_PREFERRED_URL_SCHEME'],
            TOCKY_QUEUE_DB_PATH=Path(os.environ['TOCKY_QUEUE_DB_PATH']),
        )



from typing import Literal
from tocky.utils import PageScan


def ocr_djvu_page(page_scan: PageScan, engine: Literal['easyocr', 'tesseract', 'azure'] = 'easyocr') -> str:
  if engine == 'easyocr':
    from tocky.ocr.easyocr import ocr_djvu_page_easyocr
    return ocr_djvu_page_easyocr(page_scan)
  elif engine == 'tesseract':
    from tocky.ocr.tesseract import ocr_djvu_page_tesseract
    return ocr_djvu_page_tesseract(page_scan)
  elif engine == 'azure':
    from tocky.ocr.azure import ocr_djvu_page_azure
    return ocr_djvu_page_azure(page_scan)
  else:
    raise ValueError(f'Unknown OCR engine {engine}')



from dataclasses import dataclass
from typing import Literal

from openai import OpenAI
from tocky.detector.ai_vision_detector import concatenate_and_resize, image_to_base64
from tocky.extractor import AbstractExtractor, TocEntry, TocResponse
from tocky.extractor.formats import build_system_prompt, process_extracted_output
from tocky.utils.ia import get_book_images

SYSTEM_PROMPT = """
You are a bot that helps to extract the full table of contents data in a structured format.

{format_instructions}

Notes:
- The label is used for unimportant data like numerals.
- Don't output text in ALL CAPS.

### Examples:

{PROMPT_SAMPLES[no_titles]}

{PROMPT_SAMPLES[nested]}
"""

@dataclass
class AiVisionExtractorOptions:
    model: str = "gpt-4o-mini"
    target_height: int = 512
    extraction_format: Literal['json', 'markdown'] = 'json'


class AiVisionExtractor(AbstractExtractor[AiVisionExtractorOptions]):
    name = 'ai_vision_extractor'

    def __init__(self):
        super().__init__()
        self.P = AiVisionExtractorOptions()

    def extract(self, ocaid: str, detector_result: list[int]) -> list[TocEntry]:
        toc_page_image = concatenate_and_resize(list(get_book_images(ocaid, detector_result, reduce=1)), target_height=512)

        client = OpenAI()
        system_prompt = build_system_prompt(SYSTEM_PROMPT, self.P.extraction_format)
        completion = client.chat.completions.create(
            model=self.P.model,
            messages=[
                {
                    "role": "system",
                    "content": system_prompt,
                },
                {
                    "role": "user",
                    "content": [
                        {
                            "type": "text",
                            "text": "Please extract the table of contents from this image.",
                        },
                        {
                            "type": "image_url",
                            "image_url": {
                                "url": f"data:image/png;base64,{image_to_base64(toc_page_image)}"
                            },
                        },
                    ],
                },
            ],
            # max_tokens=4096,
        )

        assert completion.choices[0].message.content
        assert completion.usage

        toc = process_extracted_output(
            completion.choices[0].message.content,
            self.P.extraction_format,
        )
        self.toc_response = TocResponse(
            toc,
            prompt_tokens=completion.usage.prompt_tokens,
            completion_tokens=completion.usage.completion_tokens,
        )

        return self.toc_response.toc



from tocky.detector import AbstractDetector
from tocky.detector.ai_vision_detector import AiVisionDetector
from tocky.detector.manual_detector import ManualDetector
from tocky.detector.ocr_detector import OcrDetector
from tocky.extractor import AbstractExtractor
from tocky.extractor.ai_extractor import AiExtractor
from tocky.extractor.ai_vision_extractor import AiVisionExtractor


DETECTERS: list[type[AbstractDetector]] = [
    OcrDetector,
    AiVisionDetector,
    ManualDetector,
]

EXTRACTORS: list[type[AbstractExtractor]] = [
    AiExtractor,
    AiVisionExtractor,
]

DETECTERS_BY_NAME = {d.name: d for d in DETECTERS}
EXTRACTORS_BY_NAME = {e.name: e for e in EXTRACTORS}



from typing import Generic, TypeVar
from tocky.utils import ShareableState

TParams = TypeVar("TParams")

class AbstractDetector(Generic[TParams]):
    name: str
    P: TParams
    S: ShareableState
    debug = True
    """
    When debug is set to true, extra helper variables could be set
    """

    def __init__(self):
        self.S = ShareableState()

    def predict_cost(self) -> float:
        raise NotImplementedError()

    def detect(self, ocaid: str) -> list[int]:
        raise NotImplementedError()



from dataclasses import dataclass, field

from tocky.detector import AbstractDetector


@dataclass
class ManualDetectorOptions:
    leaf_numbers: list[int] = field(default_factory=list)


class ManualDetector(AbstractDetector[ManualDetectorOptions]):
    name = 'manual_detector'

    def __init__(self):
        super().__init__()
        self.P = ManualDetectorOptions()

    def predict_cost(self):
        return 0.0

    def detect(self, ocaid: str):
        return self.P.leaf_numbers



import math
from typing import Literal, cast
from lxml import etree
from rtree import index

from tocky.utils.ia import get_ia_metadata

def ocr_printer_linear(djvu_page: str | etree._Element) -> str:
    if isinstance(djvu_page, str):
       root = etree.fromstring(djvu_page)
    else:
        root = djvu_page

    result = ""

    ref_word = next(
        word
        for word in root.findall('.//WORD')
        if len(word.text) > 8
    )
    ref_word_coords = ref_word.xpath("./@coords")[0].split(',')
    ref_word_width = int(ref_word_coords[2]) - int(ref_word_coords[0])
    char_width = ref_word_width / len(ref_word.text)

    for pagecol in root.findall('.//PAGECOLUMN'):
        for par in pagecol.findall('.//PARAGRAPH'):
            for line in par.findall('.//LINE'):
                words = line.findall('.//WORD')
                if words:
                    coords = line.xpath('.//WORD/@coords')[0].split(',')

                    line_indent = math.floor(int(coords[0]) / char_width)
                    line_text = ' '.join(word.text.strip() for word in words)
                    indented_line = ' ' * line_indent + line_text
                    result += indented_line + '\n'

    return result

def ocr_printer_canvas(djvu_page: str | etree._Element) -> str:
    if isinstance(djvu_page, str):
       root = etree.fromstring(djvu_page)
    else:
        root = djvu_page
    canvas = ""

    # Convert inches to pixels
    dpi = int((root.xpath(".//PARAM[@name='DPI']/@value") or [int(get_ia_metadata(ocaid)['metadata']['ppi'])])[0])
    line_rounding_size = round(0.0333 * dpi)
    img_width = int(root.xpath('./@width')[0])
    img_height = int(root.xpath('./@height')[0])

    canvas_width = 120  # dpi // 4
    conv_factor = canvas_width / img_width
    canvas_height = math.ceil(img_height * conv_factor)
    canvas = ((" " * canvas_width + '\n') * canvas_height)[:-1]

    def build_index(conv_factor: float, line_rounding_size: int) -> index.Index:
        regions = index.Index()
        rect_index = 0
        for pagecol in root.findall('.//PAGECOLUMN'):
            for par in pagecol.findall('.//PARAGRAPH'):
                for line in par.findall('.//LINE'):
                    # log('LINE')
                    line_row = None
                    for word in line.findall('.//WORD'):
                        coords = word.xpath('./@coords')[0].split(',')
                        left = int(coords[0])
                        bottom = int(coords[1])
                        right = int(coords[2])
                        top = int(coords[3])

                        baseline = round((top + (bottom - top) / 3) / line_rounding_size) * line_rounding_size
                        line_row = math.floor(baseline * conv_factor) if line_row is None else line_row
                        col = math.floor(left * conv_factor)

                        while list(regions.intersection(rect := (col, line_row, col + len(word.text), line_row + 1))):
                            col += 1

                        regions.insert(rect_index, rect, obj=word.text)
                        rect_index += 1

        return regions


    def build_canvas(regions: index.Index, canvas_width: int, canvas_height: int) -> str:
        canvas = ((" " * canvas_width + '\n') * canvas_height)[:-1]
        for item in regions.intersection(regions.bounds, objects=True):
            word = cast(str, item.object)
            line_row = int(item.bbox[1])
            col = int(item.bbox[0])

            start_idx = (canvas_width + 1) * line_row + col
            end_idx = start_idx + len(word)
            # shift = 0
            # max_shift = 5
            # log(f'{line_row}:{col}\t{coords}\t{word.text}')
            # while not canvas[max(0, start_idx - 1):(end_idx + 1)].isspace():
            #   start_idx += 1
            #   end_idx += 1
            #   shift += 1
            #   if shift >= max_shift:
            #     break
            canvas = canvas[0:start_idx] + word + canvas[end_idx:]
        return canvas
       

    def crop_text_canvas(canvas: str, used_regions: index.Index) -> str:
      lines = canvas.split('\n')
      left, bottom, right, top = list(map(int, used_regions.bounds))

      return '\n'.join((
          line[left:right] if not line.isspace() else ''
          for line in lines[bottom:top]
      ))

    def collapse_newlines(s: str):
        import re
        min_newline_chain = min(map(len, re.findall(r'\n+', s.strip(), re.MULTILINE)), default=1)

        return re.sub(r'\n' * (min_newline_chain -1) + '(\n*)', r'\1', s, flags=re.MULTILINE)
        # return re.sub(r'\n+', '\n', s, flags=re.MULTILINE)

    def collapse_spaces(s: str):
        import re
        min_newline_chain = min(map(len, re.findall(r' +', s.strip(), re.MULTILINE)), default=1)

        return re.sub(r' ' * (min_newline_chain -1) + '( *)', r'\1', s, flags=re.MULTILINE)
        # return re.sub(r'(\S) +', r'\1 ', s)

    def crunch_solos(regions: index.Index):
        def shrink_rect(rect: tuple[float, float, float, float]) -> tuple[float, float, float, float]:
            return (rect[0] + 0.1, rect[1] + 0.1, rect[2] - 0.1, rect[3] - 0.1)

        prev_rects_on_line = []
        for row_i in range(int(regions.bounds[1]), int(regions.bounds[3] + 1)):
            row_rect = (0, row_i + 0.1, canvas_width, row_i + 0.9)
            rects_on_line = list(n for n in regions.intersection(row_rect, objects=True))

            # Can I move up?
            if len(rects_on_line) == 1 and prev_rects_on_line:
                item = rects_on_line[0]
                rect = item.bbox
                new_rect = (rect[0], rect[1] - 1, rect[2], rect[3] - 1)
                
                if not any(regions.intersection(shrink_rect(new_rect))):
                    regions.delete(item.id, rect)
                    regions.insert(item.id, new_rect, obj=item.object)

                    rects_on_line = list(n for n in regions.intersection(row_rect, objects=True))
            # Can I pull predecessor?
            elif len(prev_rects_on_line) == 1 and rects_on_line:
                item = prev_rects_on_line[0]
                rect = item.bbox
                new_rect = (rect[0], rect[1] + 1, rect[2], rect[3] + 1)
                if not any(regions.intersection(shrink_rect(new_rect))):
                    regions.delete(item.id, rect)
                    regions.insert(item.id, new_rect, obj=item.object)

                    rects_on_line = list(n for n in regions.intersection(row_rect, objects=True))

            prev_rects_on_line = rects_on_line


    regions = build_index(conv_factor, line_rounding_size)
    crunch_solos(regions)
    canvas = build_canvas(regions, canvas_width, canvas_height)

    return collapse_newlines(crop_text_canvas(canvas, regions))

def print_ocr(djvu_xml: str | etree._Element, printer: Literal['canvas', 'linear']='canvas') -> str:
  if printer == 'canvas':
    return ocr_printer_canvas(djvu_xml)
  elif printer == 'linear':
    return ocr_printer_linear(djvu_xml)
  else:
    raise ValueError(f'Invalid printer type: {printer}')